#!/usr/bin/env python3

# Script makes use of pcb2gcode
# Version 1.14 of the Raspbian repositories was buggy
# Install latest version (tested with 2.2.2) of pcb2fcode with:
# sudo apt-get update
# sudo apt-get install build-essential automake autoconf autoconf-archive libtool libboost-program-options-dev libgtkmm-2.4-dev gerbv git librsvg2-dev
# git clone https://github.com/pcb2gcode/pcb2gcode.git
# cd pcb2gcode
# autoreconf -fvi
# ./configure
# make
# sudo make install
#
# Make sure to start the Raspberry in commandline mode!
# Graphics mode left not enough ram for compillation
# Compillation succeeded on Pi 3B with 1GB of RAM


import RPi.GPIO as GPIO
import time
import serial
import sys
import os
import subprocess
import math
from termios import tcflush, TCIOFLUSH


# If set to 1, the script processed Gerber files
# if set to 0 the script generates gcode files of given shapes
process_gerber = 0

process_routing = 0

process_drilling = 1

get_outline_max = 1


# If set to 1, Gcode is not sent to CNC but written in file(s)
write_2_file = 0

# Script generates sample gcode files
create_sample_files = 0
# Sample files shape size in millimeters
shape_size = 5.0

# Define max/min values for X/Y if process_gerber is not 1
# else the values will be extracted from outline.ngc
X_max = shape_size
X_min = 0.0
Y_max = 0.0
Y_min = -shape_size

# Spindle height parameters
Z_up = 5.000000
Z_down = -0.250000
Z_down_drill = -3.000000

# Feed parameters
F_slow = 100
F_fast = 3000
F_drill = 50

# Skip levelling procedure if set to 1
# run levelling if set to 0
skip_level = 1

# After leveling, the scripts can do a test run:
# With the tip of the router bit close to the surface, the CNC drives
# an elliptic curve to the maximum points.
# Test is skipped if set to 0
ellipse_test = 0
ellipse_test_routing = 0


# Route outline (cut edges) in current version just for test purposes
# while writing gcode to file
route_outline = 0

# Line ending in files must be "cusor return" + "new line" for
# external module on CNC3018pro to work
line_ending = "\r\n"

# Laser engrave graphics on top layer
# If set to 1, all actions on bottom layer (e.g. leveling) are skipped!
laser_only = 1

if(laser_only == 1):
  skip_level = 1
  process_routing = 0
  process_drilling = 0


gcode_file = "front.ngc"


level_points_X = 1
level_points_Y = 1
level_Z = [[0.0], [0.0]]
level_space_x = 20.0
level_space_y = 20.0
level_no_x = 0
level_no_y = 0

write_Z_up = " Z" + "{:.6f}".format(Z_up)
write_Z_down = "Z" + "{:.6f}".format(Z_down)
write_Z_down_drill = "Z" + "{:.6f}".format(Z_down_drill)
write_F_slow = " F" + "{:.6f}".format(F_slow)
write_F_fast = " F" + "{:.6f}".format(F_fast)


# Open USB port to CNC machine
ser = serial.Serial(
  port='/dev/ttyUSB0', #Replace ttyS0 with ttyAM0 for Pi1,Pi2,Pi0
  baudrate = 115200,
  parity=serial.PARITY_NONE,
  stopbits=serial.STOPBITS_ONE,
  bytesize=serial.EIGHTBITS,
  timeout=1
)
time.sleep(4.002)# Give port some time to really open...


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Sub process Z-level parameters
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def get_level_Z(lX, lY):
  new_Z = 0.0

  ix = 0
  lower_X = X_min
  while (lX > lower_X ):
    lower_X = lower_X + level_space_x
    ix = ix + 1
  lower_X = lower_X - level_space_x
  ix = ix - 1
  while(ix >= level_points_X-1):
    lower_X = lower_X - level_space_x
    ix = ix - 1
  higher_X = lower_X + level_space_x
  if(higher_X > X_max):
    higher_X = X_max

  iy = 0
  lower_Y = Y_min
  while (lY >  lower_Y ):
    lower_Y = lower_Y + level_space_y
    iy = iy + 1
  lower_Y = lower_Y - level_space_y
  iy = iy - 1
  while(iy >= level_points_Y-1):
    lower_Y = lower_Y - level_space_y
    iy = iy - 1
  higher_Y = lower_Y + level_space_y
  if(higher_Y > Y_max):
    higher_Y = Y_max

  factor_X = (lX - lower_X) / level_space_x
  factor_Y = (lY - lower_Y) / level_space_y

  z_diff_x1 = level_Z[ix+1][iy] - level_Z[ix][iy]
  z_diff_x2 = level_Z[ix+1][iy+1] - level_Z[ix][iy+1]

  new_z1 = level_Z[ix][iy] + z_diff_x1 * factor_X
  new_z2 = level_Z[ix][iy+1] + z_diff_x2 * factor_X

  z_diff_y1 = new_z2 - new_z1

  new_Z = new_z1 + z_diff_y1 * factor_Y

  print ("lower_X = " + str(lower_X) + ", lX = " + str(lX) + ", higher_X = " + str(higher_X))
  print ("lower_Y = " + str(lower_Y) + ", lY = " + str(lY) + ", higher_Y = " + str(higher_Y))
  print ("factor_X = " + str(factor_X) + ", factor_Y = " + str(factor_Y))
  print ("z_diff_x1 = " + str(z_diff_x1) + ", z_diff_x2 = " + str(z_diff_x2))
  print ("new_z1 = " + str(new_z1) + ", new_z2 = " + str(new_z2))
  print ("z_diff_y1 = " + str(z_diff_y1))
  print ("ix = " + str(ix) + ", iy = " + str(iy))
  print ("level_Z[ix][iy] = " + str(level_Z[ix][iy]) + ", level_Z[ix+1][iy] = " + str(level_Z[ix+1][iy]) + ", level_Z[ix][iy+1] = " + str(level_Z[ix][iy+1]) + ", level_Z[ix+1][iy+1] = " + str(level_Z[ix+1][iy+1]))
  print ("new_Z = " + str(new_Z))

  return new_Z


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Sub write line to CNC (grbl)
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def write2cnc(command_line = "", wait_for_idle = 1):

  # Empty read buffer
  ser.flushInput()
#  while(ser.read() != b''):
#    time.sleep(0.002)

  command_line = command_line + '\r' + '\n'
  ser.write(command_line.encode('utf-8'))# Write line to USB
  grbl_line = b""

  while(grbl_line != b'OK'):
    grbl_line = b""
    read_from_cnc = ser.read(1)
    while(read_from_cnc.upper() != b'\n'):
      if(read_from_cnc != b'\r'):
        grbl_line = grbl_line + read_from_cnc.upper()
      read_from_cnc = ser.read(1)

  #print ("Command " + str(command_line) + ": " + str(grbl_line) + " from Arduino")

  if(wait_for_idle == 1):
    grbl_line = b""
    while(grbl_line.find(b"IDLE") == -1):
      #print(grbl_line)
      grbl_line = b""
      read_from_cnc = ser.read(1)
      if (read_from_cnc == b''):
        ser.write(b"?\n")# Request status report
      while(read_from_cnc.upper() != b'\n'):
        if(read_from_cnc != b'\r'):
          grbl_line = grbl_line + read_from_cnc.upper()
        read_from_cnc = ser.read(1)
      time.sleep(0.002)

  # Empty read buffer
  ser.flushInput()



########################################################################
########################################################################
# Process KiCAD files if process_gerber is set to 1 (see top of file)
########################################################################
########################################################################



if(process_gerber == 1):

  #gbr2ngc_command = "pcb2gcode --back H-bridge-B_Cu.gbr --metric --zsafe 5 --zchange 100 --zwork -0.21 --offset 0.02 --mill-feed 100 --mill-speed 600  --drill H-bridge-NPTH.drl --zdrill -5 --drill-feed 100 --drill-speed 6000"

  gbr2ngc_command = "pcb2gcode --metric --front H-bridge-F_Cu.gbr --back H-bridge-B_SilkS.gbr --zsafe 5 --zchange 5 --zwork -0.21 --offset 0.525 --mill-feed 100 --mill-speed 600 --drill H-bridge-PTH.drl --zdrill -3 --drill-feed 100 --drill-speed 200 --outline H-bridge-Edge_Cuts.gbr --zcut -3 --cutter-diameter 0.001 --cut-feed 100 --cut-speed 200 --cut-infeed 100"


  rmChild = subprocess.Popen("rm " + gcode_file, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
  rmChild.wait()


  gbrChild = subprocess.Popen(gbr2ngc_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
  returnCode = 0
  for line in gbrChild.stdout.readlines():
    print (line)
  gbrChild.wait()
  returnCode = gbrChild.returncode

  print ("returnCode = " + str(returnCode) + "\n")

  if(returnCode != 0):
    print ("Error " + str(returnCode) + " generating Gcode file!\n");
    sys.exit()


if(get_outline_max == 1):
  # Get min and max X/Y from outline file ++++++++++++++++++++++++++++++++
  file_outlines = open("outline.ngc", 'r')
  file_content = file_outlines.readlines()


  string_X = "-999999.000000";
  string_Y = "-999999.000000";
  string_G = "-1";
  X_outline = -999999.0
  Y_outline = -999999.0
  G_outline = -1
  X_max = -99999
  X_min =  99999
  Y_max = -99999
  Y_min =  99999

  count = 0
  for line in file_content:
    count += 1
    #if (count < 50):
    line_temp = line.upper()[:len(line)-1] + " "
    if(line.find(";") >= 0):
      line_temp = line[:line.find(";")].upper() + " "
    if(line.find("(") >= 0):
      line_temp = line[:line.find("(")].upper() + " "

    if(line_temp.find("X") >= 0):
      string_X = line_temp[line_temp.find("X")+1:line_temp.find(" ",line_temp.find("X"))]
    if(line_temp.find("Y") >= 0):
      string_Y = line_temp[line_temp.find("Y")+1:line_temp.find(" ",line_temp.find("Y"))]
    if(line_temp.find("G") >= 0):
      string_G = line_temp[line_temp.find("G")+1:line_temp.find(" ",line_temp.find("G"))]

    X_outline = float(string_X)
    Y_outline = float(string_Y)
    G_outline = int(string_G)


    string_X = "-999999.000000";
    string_Y = "-999999.000000";
    string_G = "-1"

    if( G_outline == 20):
      G_outline = 21
      unit_conversion = 25.4

    if((X_outline < X_min) & (X_outline > -999999.0)):
      X_min = X_outline
    if(X_outline > X_max):
      X_max = X_outline
    if((Y_outline < Y_min) & (Y_outline > -999999.0)):
      Y_min = Y_outline
    if(Y_outline > Y_max):
      Y_max = Y_outline


  file_outlines.close()

  X_min = X_min * unit_conversion
  X_max = X_max * unit_conversion

  Y_min = Y_min * unit_conversion
  Y_max = Y_max * unit_conversion





########################################################################
########################################################################
# Level the surface ++++++++++++++++++++++++++++++++++++++++++++++++++++
########################################################################
########################################################################



#X_min = 0.0# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#X_max = 30.0# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#Y_min = -20.0# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#Y_max = 0.0# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


input_button = 8 # Pin header numbering P1-8
GPIO.setmode(GPIO.BOARD)
GPIO.setup(input_button, GPIO.IN, pull_up_down=GPIO.PUD_UP)

print("X_min=" + str(X_min) + ", X_max=" + str(X_max) + ", Y_min=" + str(Y_min) + ", Y_max=" + str(Y_max) + "\n")

level_points_X = round((X_max - X_min) / level_space_x) + 1

print ("level_points_X=" + str(level_points_X) + "\n")
level_space_x = (X_max - X_min) / (level_points_X - 1)

level_points_Y = round((Y_max - Y_min) / level_space_y) + 1


print ("level_points_Y=" + str(level_points_Y) + "\n")
level_space_y = (Y_max - Y_min) / (level_points_Y - 1)


level_Z = [[0 for i in range(level_points_Y)] for j in range(level_points_X)]

if(skip_level == 0):


#  grbl_line = b""
#  read_from_cnc = ser.read(1)
#  while(read_from_cnc.upper() != b'\n'):
#    if(read_from_cnc != b'\r'):
#      grbl_line = grbl_line + read_from_cnc
#    print(read_from_cnc)
#    read_from_cnc = ser.read(1)
#
#  print ("Read input ", grbl_line, " from Arduino")
#
#  grbl_line = b""
#  read_from_cnc = ser.read(1)
#  while(read_from_cnc.upper() != b'\n'):
#    if(read_from_cnc != b'\r'):
#      grbl_line = grbl_line + read_from_cnc
#    read_from_cnc = ser.read(1)
#
#  print ("Read input ", grbl_line, " from Arduino")

  print("level_space_x=" + str(level_space_x) + ", level_space_y=" + str(level_space_y))


  print("Leveling procedure starts!\n")
  print("Don't forget to connect the leveling cable to the spindle !!!\n")
  sys.stdout.flush();
  tcflush(sys.stdin, TCIOFLUSH)
  input("Press 'Enter' when ready for leveling or 'ctrl' + 'c' to exit script...")


  write2cnc("G04 P0.1", 1)# Empty grbl buffer and wait for ready signal

  write2cnc("G92 X0.000 Y0.000 Z0.000", 0)# Set current point to zero

  write2cnc("G90", 0)# Set to absolute positioning

  write2cnc("G21", 0)# Set to metric units

  # $1 sets the time in milleseconds before steppers are turned off in idle mode
  # $1=255 turns them always on
  write2cnc("$1=255", 0)# Enable all stepper motors

  write2cnc("$32=0")# Disable Laser mode

  Z_safe = 5.000000
  Z_safe_slow = 0.5

  write2cnc("Z" + "{:.6f}".format(Z_safe), 1)# Lift spindle

  #write2cnc("X10.000", 0)# Drive to coordinates
  #write2cnc("Y10.000", 0)# Drive to coordinates

  write2cnc("G01 X0.000000 Y0.000000", 1)# Drive back to zero X/Y

  #write2cnc("G04 P0.1", 1)# Empty grbl buffer before accepting next command


  y_pos = Y_min
  z_pos = 0.0
  iy = 0
  while (iy < level_points_Y):
    x_pos = X_min
    ix = 0
    while (ix < level_points_X):
      #time.sleep(1.05)
      z_pos = z_pos + Z_safe
      write_Z = " Z" + "{:.6f}".format(z_pos)
      write2cnc(write_Z, 0)# drive to high pos
      write_X = "X" + "{:.6f}".format(x_pos)
      write_Y = " Y" + "{:.6f}".format(y_pos)
      #write_Y = " Y0.000000"# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      print ("leveling point:" + write_X + " " + write_Y)
      write2cnc(write_X + write_Y, 1)# drive to next pos
      slow_mode = 0
      while slow_mode < 2:
        if(slow_mode == 0):
          z_pos = z_pos - 0.05
        else:
          z_pos = z_pos - 0.005
        write2cnc("Z" + "{:.6f}".format(z_pos),0)# Lower spindle stepwise
        if GPIO.input(input_button) == 0:
          print("Pin ", input_button, " is LOW")
          if(slow_mode == 0):
            slow_mode = 1
            z_pos = z_pos + Z_safe_slow
            write2cnc("Z" + "{:.6f}".format(z_pos), 1)# Raise spindle for second, slow run
          else:
            slow_mode = 2
            level_Z[ix][iy] = z_pos
        time.sleep(0.05)

      print("Detected contact at [" + str(ix) + "][" + str(iy) + "]: " + str(level_Z[ix][iy]))
      x_pos = x_pos + level_space_x
      ix = ix + 1
    y_pos = y_pos + level_space_y
    iy = iy + 1
    print()


  z_pos = Z_safe
  write_Z = " Z" + "{:.6f}".format(z_pos)
  write2cnc(write_Z, 1)# drive to high pos
  write2cnc("X0.000000 Y0.000000", 1)# Drive back to zero X/Y
  write2cnc("Z0.000000", 0)# drive Spindle to Zero


  if(ellipse_test == 1):
    # Test run circle around max values

    sys.stdout.flush();
    tcflush(sys.stdin, TCIOFLUSH)
    input("Leveling done, press 'Enter' when ready for test run 'ellipse'...")

    z_pos = Z_safe
    write_Z = " Z" + "{:.6f}".format(z_pos)
    write2cnc(write_Z, 0)# drive to high pos

    alpha = 0.0
    alpha_step = 0.01
    while (alpha <= 2 * math.pi + alpha_step):
      route_x = math.cos(alpha) * (X_max+X_min)/2.0 + (X_min+X_max)/2.0
      route_y = math.sin(alpha) * (Y_min+Y_max)/2.0 + (Y_min+Y_max)/2.0
      route_z = get_level_Z(route_x, route_y) + 0.20
      write_X = "X" + "{:.6f}".format(route_x)
      write_Y = " Y" + "{:.6f}".format(route_y)
      write_Z = " Z" + "{:.6f}".format(route_z)
      if(alpha == 0.0):# First coordinate, start of circle
        write_Z = "Z" + "{:.6f}".format(Z_safe)
        write2cnc(write_Z, 0)# Raise spindle to Z_safe
        write2cnc("G1" + write_X + write_Y + write_F_fast, 0)# drive to first pos
        route_z = get_level_Z(route_x, route_y) + 0.20
        write_Z = "Z" + "{:.6f}".format(route_z)
        write2cnc(write_Z, 1)# Lower Spindle
      write2cnc("G1" + write_X + write_Y + write_Z + write_F_slow, 0)# drive to next pos
      alpha = alpha + alpha_step
      #time.sleep(0.02)

    write_Z = "Z" + "{:.6f}".format(Z_safe)
    write2cnc(write_Z, 0)# Raise spindle to Z_safe
    route_x = 0.000000
    route_y = 0.000000
    write_X = "X" + "{:.6f}".format(route_x)
    write_Y = " Y" + "{:.6f}".format(route_y)
    write2cnc("G0 X0.000000 Y0.000000 F6000.000000", 0)# Zero
    write2cnc("Z0.000000 ", 1)# Spindle down to Zero

    sys.stdout.flush();
    tcflush(sys.stdin, TCIOFLUSH)
    input("Test done, press 'Enter' to continue or 'ctrl' + 'c' exit script...")

  if(ellipse_test_routing == 1):
    # Test run circle around max values with SPINDLE ON !!!

    sys.stdout.flush();
    tcflush(sys.stdin, TCIOFLUSH)
    input("Leveling done, press 'Enter' when ready for test ROUTING 'ellipse'...")

    z_pos = Z_safe
    write_Z = " Z" + "{:.6f}".format(z_pos)
    write2cnc(write_Z, 0)# drive to high pos

    ellipse_z = 0.0
    ellipse_step = -0.025
    ellipse_depth = -0.2
    input_answer = ''
    while((input_answer.upper() != 'N') & (ellipse_z > ellipse_depth)):
      alpha = 0.0
      alpha_step = 0.01
      while (alpha <= 2 * math.pi + alpha_step):
        route_x = math.cos(alpha) * (X_max+X_min)/2.0 + (X_min+X_max)/2.0
        route_y = math.sin(alpha) * (Y_min+Y_max)/2.0 + (Y_min+Y_max)/2.0
        route_z = get_level_Z(route_x, route_y) + ellipse_z
        write_X = "X" + "{:.6f}".format(route_x)
        write_Y = " Y" + "{:.6f}".format(route_y)
        write_Z = " Z" + "{:.6f}".format(route_z)
        if(alpha == 0.0):# First coordinate, start of circle
          write_Z = "Z" + "{:.6f}".format(Z_safe)
          write2cnc(write_Z+ write_F_fast, 0)# Raise spindle to Z_safe
          write2cnc("M03 S1000", 1)# Turn spindle ON
          write2cnc("G1" + write_X + write_Y + write_F_fast, 0)# drive to first pos
          route_z = get_level_Z(route_x, route_y) + ellipse_z
          write_Z = "Z" + "{:.6f}".format(route_z)
          write2cnc(write_Z+ write_F_slow, 1)# Lower Spindle
        write2cnc("G1" + write_X + write_Y + write_Z + write_F_slow, 0)# drive to next pos
        alpha = alpha + alpha_step
        #time.sleep(0.02)

      ellipse_z = ellipse_z + ellipse_step
      input_answer = ''
      while((input_answer.upper() != "Y") & (input_answer.upper() != "N")):
        input_answer = input("Do another run with lower depth (Y/N)?\n")

    write2cnc("M03 S0", 1)# Turn spindle OFF
    write_Z = "Z" + "{:.6f}".format(Z_safe)
    write2cnc(write_Z, 0)# Raise spindle to Z_safe
    route_x = 0.000000
    route_y = 0.000000
    write_X = "X" + "{:.6f}".format(route_x)
    write_Y = " Y" + "{:.6f}".format(route_y)
    write2cnc("G0 X0.000000 Y0.000000 F6000.000000", 0)# Zero
    write2cnc("Z0.000000 ", 1)# Spindle down to Zero



    sys.stdout.flush();
    tcflush(sys.stdin, TCIOFLUSH)
    input("Routing test done, press 'Enter' exit script...")

    sys.exit()# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!




########################################################################
########################################################################
# Process insulation routing file +++++++++++++++++++++++++++++++++++++
########################################################################
########################################################################

if(process_routing == 1):

  print("...and now routing starts!!!\n")
  print("Don't forget to remove the leveling cable from the spindle !!!\n")
  sys.stdout.flush();
  tcflush(sys.stdin, TCIOFLUSH)
  input("Leveling done, press 'Enter' when ready for routing or 'ctrl' + 'c' to abort script...")

  file_write = open("PCBrouteNew.ngc", "w")


  if(write_2_file == 0):
    write2cnc("G04 P0.1", 1)# Empty grbl buffer and wait for ready signal
    write2cnc("G92 X0.000 Y0.000 Z0.000", 0)# Set current point to zero
    write2cnc("G90", 0)# Set to absolute positioning
    write2cnc("G21", 0)# Set to metric units
    write2cnc("$32=0")# Disable Laser mode
  else:
    file_write.write("G04 P0.1" + line_ending)
    file_write.write("G92 X0.000 Y0.000 Z0.000" + line_ending)
    file_write.write("G90" + line_ending)
    file_write.write("G21" + line_ending)# Metric units


  # $1 sets the time in milleseconds before steppers are turned off in idle mode
  # $1=255 turns them always on
  if(write_2_file == 0):
    write2cnc("$1=255", 0)# Enable all stepper motors


  if(route_outline == 1):
    # Route outline with max/min values
    write_route_z = "Z" + "{:.6f}".format(Z_up)
    if(write_2_file == 1):
      write_route_x = "X" + "{:.6f}".format(X_min)
      write_route_y = " Y" + "{:.6f}".format(Y_min)
      write_route_z = " Z" + "{:.6f}".format(Z_up)
      write_route_f = " F" + "{:.6f}".format(F_fast)
      file_write.write(write_route_z + write_route_f + line_ending)
      file_write.write("G00" + write_route_x + write_route_y + write_route_f + line_ending)
      write_route_z = " Z" + "{:.6f}".format(Z_down)
      write_route_f = " F" + "{:.6f}".format(F_slow)
      file_write.write(write_route_z + write_route_f + line_ending)
      write_route_x = "X" + "{:.6f}".format(X_min)
      write_route_y = " Y" + "{:.6f}".format(Y_max)
      file_write.write("G01" + write_route_x + write_route_y + write_route_z + write_route_f + line_ending)
      write_route_x = "X" + "{:.6f}".format(X_max)
      write_route_y = " Y" + "{:.6f}".format(Y_max)
      file_write.write("G01" + write_route_x + write_route_y + write_route_z + write_route_f + line_ending)
      write_route_x = "X" + "{:.6f}".format(X_max)
      write_route_y = " Y" + "{:.6f}".format(Y_min)
      file_write.write("G01" + write_route_x + write_route_y + write_route_z + write_route_f + line_ending)
      write_route_x = "X" + "{:.6f}".format(X_min)
      write_route_y = " Y" + "{:.6f}".format(Y_min)
      file_write.write("G01" + write_route_x + write_route_y + write_route_z + write_route_f + line_ending)
      write_route_z = " Z" + "{:.6f}".format(Z_up)
      write_route_f = " F" + "{:.6f}".format(F_fast)
      file_write.write(write_route_z + write_route_f + line_ending)
      write_route_x = "X" + "{:.6f}".format(0.000000)
      write_route_y = " Y" + "{:.6f}".format(0.000000)
      file_write.write("G01" + write_route_x + write_route_y + write_route_f + line_ending)


  file_read = open(gcode_file, 'r')
  file_content = file_read.readlines()

  # Read file

  Z_down = -0.0
  Z_down_step = -0.025
  Z_down_max = -0.45
  input_answer = ''
  while((input_answer.upper() != 'N') & (Z_down > Z_down_max)):

    string_X = "-999999.000000";
    string_Y = "-999999.000000";
    string_Z = "-999999.000000";
    string_F = "-1.000000";
    string_P = "-1.000000";
    P = [-1]
    X = [-999999.0]
    Y = [-999999.0]
    Z = [-999999.0]
    F = [-1]
    #X_max = -99999
    #X_min =  99999
    #Y_max = -99999
    #Y_min =  99999
    Z_max = -99999
    Z_min =  99999

    string_G = "-1"
    G = [-1]

    write_route_x = "X" + "{:.6f}".format(X_min)
    write_route_y = " Y" + "{:.6f}".format(Y_min)
    write_route_z = " Z" + "{:.6f}".format(Z_up)
    write_route_f = " F" + "{:.6f}".format(F_fast)

    unit_conversion = 1.0

    route_x = 0.000000
    route_x_old = 0.000000
    route_y = 0.000000
    route_y_old = 0.000000
    route_z = Z_up
    route_z_old = Z_up
    route_g = 0
    route_g_old = 0
    route_f = 0.000000
    route_f_old = 0.000000

    string_X = "0.000000";
    string_Y = "0.000000";
    string_Z = "0.000000";
    string_F = str(F_fast);
    string_P = "0.000000";
    string_G = "-1"


    # Raise spindle
    write_route_z = "Z" + "{:.6f}".format(Z_up)
    if(write_2_file == 0):
      write2cnc(write_route_z, 1)# Raise spindle to Z_safe
    else:
      file_write.write("G01 " + write_route_x + write_route_y + write_route_z + write_route_f + line_ending)

    # Turn spindle ON
    if(write_2_file == 0):
      write2cnc("M03 S1000", 1)
    else:
      file_write.write("M03 S1000" + line_ending)

    count = 0
    for line in file_content:
      count += 1
      if(count > -1):
        line_temp = line.upper()[:len(line)-1] + " "
        if(line.find(";") >= 0):
          line_temp = line[:line.find(";")].upper() + " "
        if(line.find("(") >= 0):
          line_temp = line[:line.find("(")].upper() + " "

        if(line_temp.find("X") >= 0):
          string_X = line_temp[line_temp.find("X")+1:line_temp.find(" ",line_temp.find("X"))]
        if(line_temp.find("Y") >= 0):
          string_Y = line_temp[line_temp.find("Y")+1:line_temp.find(" ",line_temp.find("Y"))]
        if(line_temp.find("Z") >= 0):
          string_Z = line_temp[line_temp.find("Z")+1:line_temp.find(" ",line_temp.find("Z"))]
        if(line_temp.find("F") >= 0):
          string_F = line_temp[line_temp.find("F")+1:line_temp.find(" ",line_temp.find("F"))]
        if(line_temp.find("P") >= 0):
          string_P = line_temp[line_temp.find("P")+1:line_temp.find(" ",line_temp.find("P"))]
        if(line_temp.find("G") >= 0):
          string_G = line_temp[line_temp.find("G")+1:line_temp.find(" ",line_temp.find("G"))]

        #print("file_content{}: {}".format(count, line.strip()) + " string_X = " + string_X)
        #print("X=" + string_X + ", Y=" + string_Y + ", Z=" + string_Z)

        route_g = int(string_G)
        if( route_g == 20):
          route_g = 21
          unit_conversion = 25.4

        route_x = float(string_X) * unit_conversion
        route_y = float(string_Y) * unit_conversion
        route_z = float(string_Z) * unit_conversion
        if(route_z < 0.0):
          route_z = Z_down
          route_f = F_slow
        else:
          route_z = Z_up
          route_f = F_fast

        route_p = float(string_P)
        route_g = int(string_G)


        if((route_x != route_x_old) | (route_y != route_y_old)):
          distance = math.sqrt(pow((route_x - route_x_old), 2) + pow((route_y - route_y_old), 2))
          # Insert extra points if distance between two points along path is too large
          # Thats to make sure level points are not ignored over long distances
          # Only needed when spindle is down
          distance_max = 10.0
          while((distance > distance_max) & (route_z == Z_down) & (route_z_old == Z_down)):
            insert_x = route_x_old + (route_x - route_x_old) * distance_max / distance
            insert_y = route_y_old + (route_y - route_y_old) * distance_max / distance
            insert_z = get_level_Z(insert_x, insert_y) + Z_down
            write_route_x = "X" + "{:.6f}".format(insert_x)
            write_route_y = " Y" + "{:.6f}".format(insert_y)
            write_route_z = " Z" + "{:.6f}".format(insert_z)
            write_route_f = " F" + "{:.6f}".format(F_slow)
            if(write_2_file == 0):
              write2cnc("G01 " + write_route_x + write_route_y + write_route_z + write_route_f, 0)
            else:
              file_write.write("G01 " + write_route_x + write_route_y + write_route_z + write_route_f + line_ending)
            route_x_old = insert_x
            route_y_old = insert_y
            distance = math.sqrt(pow((route_x - route_x_old), 2) + pow((route_y - route_y_old), 2))


        if(route_x != route_x_old):
          write_route_x = " X" + "{:.6f}".format(route_x)
        else:
          write_route_x = ""

        if(route_y != route_y_old):
          write_route_y = " Y" + "{:.6f}".format(route_y)
        else:
          write_route_y = ""

        if(route_z != route_z_old):
          write_route_z = " Z" + "{:.6f}".format(route_z)
        else:
          write_route_z = ""

        if(route_f != route_f_old):
          write_route_f = " F" + "{:.6f}".format(route_f)
        else:
          write_route_f = ""

        if((route_x != route_x_old) | (route_y != route_y_old) | (route_z != route_z_old)):
          if((route_z == Z_up) & (route_z_old == Z_down)):
            write_route_z = " Z" + "{:.6f}".format(route_z)
            if(write_2_file == 0):
              write2cnc(write_route_z + write_route_f, 0)
            else:
              file_write.write(write_route_z + write_route_f + line_ending)

          if((route_z_old == Z_up) & (route_z == Z_down)):
            route_z = get_level_Z(route_x, route_y) + Z_down
            write_route_z = " Z" + "{:.6f}".format(route_z)
            if(write_2_file == 0):
              write2cnc(write_route_z + write_route_f, 0)
            else:
              file_write.write(write_route_z + write_route_f + line_ending)

          route_x_old = route_x
          route_y_old = route_y
          route_z_old = route_z
          if(route_z == Z_down):
            route_z = get_level_Z(route_x, route_y) + Z_down
            write_route_z = " Z" + "{:.6f}".format(route_z)

          if(write_2_file == 0):
            write2cnc("G01" + write_route_x + write_route_y + write_route_z + write_route_f, 0)
          else:
            file_write.write("G01" + write_route_x + write_route_y + write_route_z + write_route_f + line_ending)

      else:# if count > 92
        break

    write_route_z = "Z" + "{:.6f}".format(Z_up)
    if(write_2_file == 0):
      write2cnc(write_route_z, 0)# Raise spindle to Z_safe
    else:
      file_write.write("G01 " + write_route_x + write_route_y + write_route_z + write_route_f + line_ending)

    if(write_2_file == 0):
      write2cnc("M03 S0", 1)# Turn spindle OFF
      write2cnc("G01 X0.000000 Y0.000000" + write_F_fast, 1)# Back to Zero
      write2cnc("Z0.000000", 1)# Spindle down to Zero
    else:
      file_write.write("M03 S0" + line_ending)# Turn spindle OFF
      file_write.write("G01 X0.000000 Y0.000000" + write_F_fast + line_ending)# Back to Zero
      file_write.write("Z0.000000" + line_ending)# Spindle down to Zero

    print("X_min=" + str(X_min) + ", X_max=" + str(X_max) + ", Y_min=" + str(Y_min) + ", Y_max=" + str(Y_max) + ", Z_min=" + str(Z_min) + ", Z_max=" + str(Z_max))

    Z_down = Z_down + Z_down_step
    input_answer = ''
    while((input_answer.upper() != "Y") & (input_answer.upper() != "N")):
      sys.stdout.flush();
      tcflush(sys.stdin, TCIOFLUSH)
      input_answer = input("Do another run with lower depth (Y/N)?\n")



  file_write.close()
  file_read.close()


########################################################################
########################################################################
# Process drill file +++++++++++++++++++++++++++++++++++++++++++++++++++
########################################################################
########################################################################

if(process_drilling == 1):

  print("Routing done, now prepare for drilling!!!\n")
  print("Replace V-bit with drill and drive drill close to board surface !!!\n")
  sys.stdout.flush();
  tcflush(sys.stdin, TCIOFLUSH)
  input("Press 'Enter' when ready for drilling or 'ctrl' + 'c' to abort script...")

  # Generate Gcode drill file ++++++++++++++++++++++++++++++++++++++++++++
  file_read = open("H-bridge-PTH.drl", 'r')
  file_content = file_read.readlines()

  string_X = "0.000000";
  string_Y = "0.000000";
  string_Z = "0.000000";
  string_F = "0.000000";
  string_P = "0.000000";
  X = 0
  Y = 0
  Z = 0
  F = 0
  P = 0
  string_G = "0"
  G = 0

  # Write processed data into new file
  if(write_2_file == 0):
    write2cnc("G92", 1)# Make sure current position is Zero
    write2cnc("G90", 0)# Absolute positioning
    write2cnc("G21", 0)# Metric units
    write2cnc("$32=0")# Disable Laser mode
  else:
    file_write = open("PCBdrillNew.ngc", "w")
    file_write.write("G92" + line_ending)# Make sure current position is Zero
    file_write.write("G90" + line_ending)# Absolute positioning
    file_write.write("G21" + line_ending)# Metric units


  # Drill holes on corner points
  if(1 == 1):
    write_route_x = "X" + "{:.6f}".format(X_min)
    write_route_y = " Y" + "{:.6f}".format(Y_min)
    write_route_z = " Z" + "{:.6f}".format(Z_up)
    write_route_f = " F" + "{:.6f}".format(F_fast)
    if(write_2_file == 0):
      write2cnc(write_route_z + write_route_f, 1)
      write2cnc("M03 S1000", 1)# Turn spindle ON
      write2cnc("G00" + write_route_x + write_route_y + write_route_f, 1)
    else:
      file_write.write(write_route_z + write_route_f + line_ending)
      file_write.write("M03 S1000" + line_ending)# Turn spindle ON
      file_write.write("G00" + write_route_x + write_route_y + write_route_f + line_ending)

    write_route_z = " Z" + "{:.6f}".format(Z_down_drill)
    write_route_f = " F" + "{:.6f}".format(F_slow)
    if(write_2_file == 0):
      write2cnc("G00" + write_route_x + write_route_y + write_route_f, 1)
      write2cnc(write_route_z + write_route_f, 1)
    else:
      file_write.write(write_route_z + write_route_f + line_ending)

    write_route_z = " Z" + "{:.6f}".format(Z_up)
    write_route_f = " F" + "{:.6f}".format(F_fast)
    if(write_2_file == 0):
      write2cnc(write_route_z + write_route_f, 1)
    else:
      file_write.write(write_route_z + write_route_f + line_ending)

    write_route_x = "X" + "{:.6f}".format(X_min)
    write_route_y = " Y" + "{:.6f}".format(Y_max)
    if(write_2_file == 0):
      write2cnc("G01 " + write_route_x + write_route_y + write_route_z + write_route_f, 0)
    else:
      file_write.write("G01 " + write_route_x + write_route_y + write_route_z + write_route_f + line_ending)
    write_route_z = " Z" + "{:.6f}".format(Z_down_drill)
    write_route_f = " F" + "{:.6f}".format(F_slow)
    if(write_2_file == 0):
      write2cnc(write_route_z + write_route_f, 1)
    else:
      file_write.write(write_route_z + write_route_f + line_ending)
    write_route_z = " Z" + "{:.6f}".format(Z_up)
    write_route_f = " F" + "{:.6f}".format(F_fast)
    if(write_2_file == 0):
      write2cnc(write_route_z + write_route_f, 1)
    else:
      file_write.write(write_route_z + write_route_f + line_ending)

    write_route_x = "X" + "{:.6f}".format(X_max)
    write_route_y = " Y" + "{:.6f}".format(Y_max)
    if(write_2_file == 0):
      write2cnc("G01 " + write_route_x + write_route_y + write_route_z + write_route_f, 0)
    else:
      file_write.write("G01 " + write_route_x + write_route_y + write_route_z + write_route_f + line_ending)
    write_route_z = " Z" + "{:.6f}".format(Z_down_drill)
    write_route_f = " F" + "{:.6f}".format(F_slow)
    if(write_2_file == 0):
      write2cnc(write_route_z + write_route_f, 1)
    else:
      file_write.write(write_route_z + write_route_f + line_ending)
    write_route_z = " Z" + "{:.6f}".format(Z_up)
    write_route_f = " F" + "{:.6f}".format(F_fast)
    if(write_2_file == 0):
      write2cnc(write_route_z + write_route_f, 1)
    else:
      file_write.write(write_route_z + write_route_f + line_ending)

    write_route_x = "X" + "{:.6f}".format(X_max)
    write_route_y = " Y" + "{:.6f}".format(Y_min)
    if(write_2_file == 0):
      write2cnc("G01 " + write_route_x + write_route_y + write_route_z + write_route_f, 0)
    else:
      file_write.write("G01 " + write_route_x + write_route_y + write_route_z + write_route_f + line_ending)
    write_route_z = " Z" + "{:.6f}".format(Z_down_drill)
    write_route_f = " F" + "{:.6f}".format(F_slow)
    if(write_2_file == 0):
      write2cnc(write_route_z + write_route_f, 1)
    else:
      file_write.write(write_route_z + write_route_f + line_ending)
    write_route_z = " Z" + "{:.6f}".format(Z_up)
    write_route_f = " F" + "{:.6f}".format(F_fast)
    if(write_2_file == 0):
      write2cnc(write_route_z + write_route_f, 1)
    else:
      file_write.write(write_route_z + write_route_f + line_ending)

    write_route_x = "X" + "{:.6f}".format(X_min)
    write_route_y = " Y" + "{:.6f}".format(Y_min)
    if(write_2_file == 0):
      write2cnc("G01 " + write_route_x + write_route_y + write_route_z + write_route_f, 0)
    else:
      file_write.write("G01 " + write_route_x + write_route_y + write_route_z + write_route_f + line_ending)
    write_route_z = " Z" + "{:.6f}".format(Z_down_drill)
    write_route_f = " F" + "{:.6f}".format(F_slow)
    if(write_2_file == 0):
      write2cnc(write_route_z + write_route_f, 1)
    else:
      file_write.write(write_route_z + write_route_f + line_ending)
    write_route_z = " Z" + "{:.6f}".format(Z_up)
    write_route_f = " F" + "{:.6f}".format(F_fast)
    if(write_2_file == 0):
      write2cnc(write_route_z + write_route_f, 1)
    else:
      file_write.write(write_route_z + write_route_f + line_ending)

    write_route_z = " Z" + "{:.6f}".format(Z_up)
    write_route_f = " F" + "{:.6f}".format(F_fast)
    if(write_2_file == 0):
      write2cnc(write_route_z + write_route_f, 1)
    else:
      file_write.write(write_route_z + write_route_f + line_ending)
    write_route_x = "X" + "{:.6f}".format(0.000000)
    write_route_y = " Y" + "{:.6f}".format(0.000000)
    if(write_2_file == 0):
      write2cnc("G01 " + write_route_x + write_route_y + write_route_z + write_route_f, 0)
    else:
      file_write.write("G01 " + write_route_x + write_route_y + write_route_z + write_route_f + line_ending)


  unit_conversion = 1.0
  F = 100.000000
  for line in file_content:
    line_temp = line.upper()[:len(line)-1] + " "
    if(line.find(";") >= 0):
      line_temp = line[:line.find(";")].upper() + " "
    if(line.find("(") >= 0):
      line_temp = line[:line.find("(")].upper() + " "

    if(line_temp.find("X") >= 0):
      string_X = line_temp[line_temp.find("X")+1:line_temp.find("Y",line_temp.find("X"))]
    if(line_temp.find("Y") >= 0):
      string_Y = line_temp[line_temp.find("Y")+1:line_temp.find(" ",line_temp.find("Y"))]


    if(line_temp.find("X") >= 0 & line_temp.find("Y") >= 0):
      X = (float(string_X) - X_min*0) * unit_conversion
      Y = (float(string_Y) - Y_min*0) * unit_conversion
      write_X = ""
      write_Y = ""
      if(line_temp.find("X") >= 0):
        write_X = " X" + "{:.6f}".format(X)
      if(line_temp.find("Y") >= 0):
        write_Y = " Y" + "{:.6f}".format(Y)

      if(write_2_file == 0):
        write2cnc(write_Z_up + " " + write_F_fast, 0)
        write2cnc(write_X + write_Y, 0)
        write2cnc(write_Z_down_drill + " " + write_F_slow, 1)
        write2cnc(write_Z_up + " " + write_F_fast, 0)
      else:
        file_write.write(write_Z_up + " " + write_F_fast + line_ending)
        file_write.write(write_X + write_Y + line_ending)
        file_write.write(write_Z_down_drill + " " + write_F_slow + line_ending)


  if(write_2_file == 0):
    write2cnc(write_Z_up + " " + write_F_fast, 1)
    write2cnc("M03 S0", 1)# Turn spindle OFF
    write2cnc("X0.000000 Y0.000000", 1)
  else:
    file_write.write(write_Z_up + " " + write_F_fast + line_ending)
    file_write.write("M03 S0" + line_ending)# Turn spindle OFF
    file_write.write("X0.000000 Y0.000000" + line_ending)
    file_write.close()


  file_read.close()


  # $1 sets the time in milleseconds before steppers are turned off in idle mode
  # $1=255 turns them always on
  write2cnc("$1=50", 1)# Disable all stepper motors after 50ms idle


  print("Drilling done, Labeling starts!!!\n")
  print("Mount laser module and adjust board with NON copper layer on top !!!\n")
  sys.stdout.flush();
  tcflush(sys.stdin, TCIOFLUSH)
  input("Press 'Enter' when ready for laser marking or 'ctrl' + 'c' to abort script...")


########################################################################
########################################################################
# Laser graphics on top layer ++++++++++++++++++++++++++++++++++++++++++
########################################################################
########################################################################

if(laser_only == 1):

  print("...and now laser engraving starts!!!\n")
  print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n")
  print("Put on your goggles!!!\n")
  print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n")
  sys.stdout.flush();
  tcflush(sys.stdin, TCIOFLUSH)
  input("Install laser module and press 'Enter' when ready or 'ctrl' + 'c' to abort script...")



  write_route_z = " Z0.000000"
  F_laser = 4000
  write_route_f = " F" + "{:.6f}".format(F_laser)

  write2cnc("G04 P0.1", 1)# Empty grbl buffer and wait for ready signal
  write2cnc("G92 X0.000 Y0.000 Z0.000", 0)# Set current point to zero
  write2cnc("G90", 0)# Set to absolute positioning
  write2cnc("G21", 0)# Set to metric units
  # $1 sets the time in milleseconds before steppers are turned off in idle mode
  # $1=255 turns them always on
  write2cnc("$1=255", 0)# Enable all stepper motors

  write2cnc("$32=1")# Enable Laser mode


  # Drive to min and max points to adjust PCB alignment

  input_answer = ''
  write_laser_low = "M03 S10"
  laser_high = 270  # 1000 is fully ON
  write_laser_high = "M03 S" + "{:.6f}".format(laser_high)
  laser_gap = 0.1
  laser_gap_rest = 0.0
  laser_pause = 0.01
  write_laser_pause = "G4 P" + "{:.6f}".format(laser_pause)
  print ("write_laser_pause=" + write_laser_pause + "\n")

  while(input_answer.upper() != 'N'):
    write_route_x = "X" + "{:.6f}".format(X_min)
    write_route_y = " Y" + "{:.6f}".format(Y_min)
    write2cnc(write_route_z + write_route_f, 0)# Laser height to Zero
    write2cnc("G1 " + write_route_x + write_route_y + write_route_f, 1)
    write2cnc(write_laser_low, 0)

    sys.stdout.flush();
    tcflush(sys.stdin, TCIOFLUSH)
    print("\n\n\nAdjust Laser focus and PCB bottom left to Laser beam")
    input_answer = input("Press 'Enter' when done or 'ctrl' + 'c' to abort script...\n")

    write2cnc("M03 S0", 1)# Laser OFF
    write_route_x = "X" + "{:.6f}".format(X_max)
    write_route_y = " Y" + "{:.6f}".format(Y_max)
    write2cnc("G01 " + write_route_x + write_route_y + write_route_f, 0)
    write2cnc(write_laser_low, 1)

    sys.stdout.flush();
    tcflush(sys.stdin, TCIOFLUSH)
    print("\n\n\nNow adjust PCB top right to Laser beam")
    input_answer = input("Press 'Enter' when done or 'ctrl' + 'c' to abort script...\n")

    write2cnc("M03 S0", 1)# Laser OFF

    input_answer = ''
    while((input_answer.upper() != "Y") & (input_answer.upper() != "N")):
      sys.stdout.flush();
      tcflush(sys.stdin, TCIOFLUSH)
      input_answer = input("Do another run for laser adjustment (Y/N)?\n")



  print("Laser engraving with HIGH POWER starts!!!\n")
  print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n")
  print("Put on your goggles!!!\n")
  print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n")
  input_answer = ''
  while((input_answer.upper() != "YES") & (input_answer.upper() != "NO")):
    sys.stdout.flush();
    tcflush(sys.stdin, TCIOFLUSH)
    input_answer = input("Start laser engraving with HIGH POWER BEAM (YES / NO)?\n")

  if(input_answer.upper() == "NO"):
    sys.exit()# Leave script
  else:
    # Now process the graphics file which should be the backside
    file_read = open("back.ngc", 'r')
    file_content = file_read.readlines()

    # Read file

    input_answer = ''
    while((input_answer.upper() != 'N') & (laser_high < 1000)):

      string_X = "-999999.000000";
      string_Y = "-999999.000000";
      string_Z = "-999999.000000";
      string_F = "-1.000000";
      string_P = "-1.000000";
      P = [-1]
      X = [-999999.0]
      Y = [-999999.0]
      Z = [-999999.0]
      F = [-1]
      #X_max = -99999
      #X_min =  99999
      #Y_max = -99999
      #Y_min =  99999
      Z_max = -99999
      Z_min =  99999

      string_G = "-1"
      G = [-1]

      write_route_x = "X" + "{:.6f}".format(X_min)
      write_route_y = " Y" + "{:.6f}".format(Y_min)

      unit_conversion = 1.0

      route_x = 0.000000
      route_x_old = 0.000000
      route_y = 0.000000
      route_y_old = 0.000000
      route_z = Z_up
      route_z_old = Z_up
      route_g = 0
      route_g_old = 0
      route_f = 0.000000
      route_f_old = 0.000000

      string_X = "0.000000";
      string_Y = "0.000000";
      string_Z = "0.000000";
      string_F = str(F_fast);
      string_P = "0.000000";
      string_G = "-1"

      count = 0
      for line in file_content:
        count += 1
        if(count < 118):# Process only a given number of lines for test only
          line_temp = line.upper()[:len(line)-1] + " "
          if(line.find(";") >= 0):
            line_temp = line[:line.find(";")].upper() + " "
          if(line.find("(") >= 0):
            line_temp = line[:line.find("(")].upper() + " "

          if(line_temp.find("X") >= 0):
            string_X = line_temp[line_temp.find("X")+1:line_temp.find(" ",line_temp.find("X"))]
          if(line_temp.find("Y") >= 0):
            string_Y = line_temp[line_temp.find("Y")+1:line_temp.find(" ",line_temp.find("Y"))]
          if(line_temp.find("Z") >= 0):
            string_Z = line_temp[line_temp.find("Z")+1:line_temp.find(" ",line_temp.find("Z"))]
          if(line_temp.find("F") >= 0):
            string_F = line_temp[line_temp.find("F")+1:line_temp.find(" ",line_temp.find("F"))]
          if(line_temp.find("P") >= 0):
            string_P = line_temp[line_temp.find("P")+1:line_temp.find(" ",line_temp.find("P"))]
          if(line_temp.find("G") >= 0):
            string_G = line_temp[line_temp.find("G")+1:line_temp.find(" ",line_temp.find("G"))]

          #print("file_content{}: {}".format(count, line.strip()) + " string_X = " + string_X)
          #print("X=" + string_X + ", Y=" + string_Y + ", Z=" + string_Z)

          route_g = int(string_G)
          if( route_g == 20):
            route_g = 21
            unit_conversion = 25.4

          route_x = float(string_X) * unit_conversion + X_max
          route_y = float(string_Y) * unit_conversion
          route_z = float(string_Z) * unit_conversion

          if(route_z >= 0.0):      # Spindle up means Laser is off
            laser_gap_rest = 0.0
            route_z = Z_up
          else:
            route_z = Z_down

          route_p = float(string_P)
          route_g = int(string_G)


          if(route_x != route_x_old):
            write_route_x = " X" + "{:.6f}".format(route_x)
          else:
            write_route_x = ""

          if(route_y != route_y_old):
            write_route_y = " Y" + "{:.6f}".format(route_y)
          else:
            write_route_y = ""


          if((route_x != route_x_old) | (route_y != route_y_old) | (route_z != route_z_old)):


            distance = laser_gap_rest + math.sqrt(pow((route_x - route_x_old), 2) + pow((route_y - route_y_old), 2))
            # Insert extra points if distance between two points along path is too large
            # Thats to make sure level points are not ignored over long distances
            # Only needed when spindle is down
            print ("Distance=" + str(distance) + ", laser_gap_rest=" + str(laser_gap_rest))
            while((distance >= laser_gap) & (route_z == Z_down) & (route_z_old == Z_down)):
              insert_x = route_x_old + (route_x - route_x_old) * (laser_gap - laser_gap_rest) / distance
              insert_y = route_y_old + (route_y - route_y_old) * (laser_gap - laser_gap_rest) / distance
              write_route_x = "X" + "{:.6f}".format(insert_x)
              write_route_y = " Y" + "{:.6f}".format(insert_y)
              write_route_f = " F" + "{:.6f}".format(F_laser)
              write2cnc("G1 " + write_route_x + write_route_y, 0)
              write2cnc(write_laser_high, 0)  # Turn Laser ON
              write2cnc(write_laser_pause, 0) # Laser on for xx milliseconds
              write2cnc("M03 S0", 0)          # Turn Laser OFF
              route_x_old = insert_x
              route_y_old = insert_y
              laser_gap_rest = 0.0
              distance = math.sqrt(pow((route_x - route_x_old), 2) + pow((route_y - route_y_old), 2))

            distance = math.sqrt(pow((route_x - route_x_old), 2) + pow((route_y - route_y_old), 2))
            laser_gap_rest = laser_gap_rest + distance





            route_x_old = route_x
            route_y_old = route_y
            route_z_old = route_z

            write2cnc("G1" + write_route_x + write_route_y + write_route_f, 0)

        else:# if count > 92
          break

      write2cnc("M03 S0", 1)# Turn Laser OFF
      write2cnc("G01 X0.000000 Y0.000000" + write_F_fast, 1)# Back to Zero
      write2cnc("Z0.000000", 1)# Spindle down to Zero

      print("X_min=" + str(X_min) + ", X_max=" + str(X_max) + ", Y_min=" + str(Y_min) + ", Y_max=" + str(Y_max) + ", Z_min=" + str(Z_min) + ", Z_max=" + str(Z_max))


      laser_high = laser_high + 100  # 1000 is fully ON
      if(laser_high > 1000):
        laser_high = 1000
      write_laser_high = "M03 S" + "{:.6f}".format(laser_high)

      input_answer = ''
      while((input_answer.upper() != "Y") & (input_answer.upper() != "N")):
        input_answer = input("Do another run with higher laser power (Y/N)?\n")


    write2cnc("$32=0")# Disable Laser mode
    file_read.close()



########################################################################
########################################################################
# Create sample Gcode files if create_sample_files is set to 1
########################################################################
########################################################################
if(create_sample_files == 1):
  shape_type = 0
  while (shape_type < 3):

    file_name_gcode = "square.ngc"
    if(shape_type == 1):
      file_name_gcode = "circle.ngc"
    if(shape_type == 2):
      file_name_gcode = "octagon.ngc"

    file_write = open(file_name_gcode, "w")
    file_write.write("G90" + line_ending)# Absolute positioning
    file_write.write("G92" + line_ending)# Make sure current position is Zero
    file_write.write("G21" + line_ending)# Metric units

    file_write.write("G0" + write_Z_up + write_F_fast + line_ending)

    if(shape_type == 0):# Square
      X = X_min
      Y = Y_min
      file_write.write("G1 " + "X" + "{:.6f}".format(X) + " Y" + "{:.6f}".format(Y) + line_ending)
      file_write.write("M03 S1000" + line_ending)# Turn spindle fully ON
      file_write.write("G1 " + write_Z_down + write_F_slow + line_ending)
      X = X_min
      Y = Y_max
      file_write.write("G1 " + "X" + "{:.6f}".format(X) + " Y" + "{:.6f}".format(Y) + line_ending)
      X = X_max
      Y = Y_max
      file_write.write("G1 " + "X" + "{:.6f}".format(X) + " Y" + "{:.6f}".format(Y) + line_ending)
      X = X_max
      Y = Y_min
      file_write.write("G1 " + "X" + "{:.6f}".format(X) + " Y" + "{:.6f}".format(Y) + line_ending)
      X = X_min
      Y = Y_min
      file_write.write("G1 " + "X" + "{:.6f}".format(X) + " Y" + "{:.6f}".format(Y) + line_ending)

      file_write.write("G1 " + write_Z_up + write_F_fast + line_ending)
      file_write.write("M03 S0" + line_ending)# Turn spindle OFF
      X = 0.000000
      Y = 0.000000
      file_write.write("G0 " + "X" + "{:.6f}".format(X) + " Y" + "{:.6f}".format(Y) + line_ending)


    if(shape_type == 1):# Circle

      alpha = 0.0
      alpha_step = 0.1
      while (alpha <= 2 * math.pi + alpha_step):
        X = math.cos(alpha) * (X_max+X_min)/2.0 + (X_min+X_max)/2.0
        Y = math.sin(alpha) * (Y_min+Y_max)/2.0 + (Y_min+Y_max)/2.0
        file_write.write("G1 " + "X" + "{:.6f}".format(X) + " Y" + "{:.6f}".format(Y) + line_ending)
        if(alpha == 0.0):
          file_write.write("M03 S1000" + line_ending)# Turn spindle fully ON
          file_write.write("G1 " + write_Z_down + write_F_slow + "\n")
        alpha = alpha + alpha_step

      file_write.write("G1 " + write_Z_up + write_F_fast + line_ending)
      file_write.write("M03 S0" + line_ending)# Turn spindle OFF
      X = 0.000000
      Y = 0.000000
      file_write.write("G0 " + "X" + "{:.6f}".format(X) + " Y" + "{:.6f}".format(Y) + line_ending)


    if(shape_type == 2):# Octagon

      alpha = 0.0
      alpha_step = math.pi / 4
      while (alpha <= 2 * math.pi + alpha_step):
        X = math.cos(alpha) * (X_max+X_min)/2.0 + (X_min+X_max)/2.0
        Y = math.sin(alpha) * (Y_min+Y_max)/2.0 + (Y_min+Y_max)/2.0
        file_write.write("G1 " + "X" + "{:.6f}".format(X) + " Y" + "{:.6f}".format(Y) + line_ending)
        if(alpha == 0.0):
          file_write.write("M03 S1000" + line_ending)# Turn spindle fully ON
          file_write.write("G1 " + write_Z_down + write_F_slow + "\n")
        alpha = alpha + alpha_step

      file_write.write("G1 " + write_Z_up + write_F_fast + line_ending)
      file_write.write("M03 S0" + line_ending)# Turn spindle OFF
      X = 0.000000
      Y = 0.000000
      file_write.write("G0 " + "X" + "{:.6f}".format(X) + " Y" + "{:.6f}".format(Y) + line_ending)

    file_write.write("G1 " + "Z0.000000".format(Y) + line_ending)

    file_write.write("G04 P0" + line_ending)# Pause machine for 2 seconds
    file_write.write("M02" + line_ending)# Tell CNC this is the end of program

    file_write.close()

    shape_type = shape_type + 1



